<?php

namespace Myriad\Queue\Drivers\Google;

use Carbon\Carbon;
use DateTime;
use DateTimeInterface;
use Exception;
use Google\Cloud\PubSub\PubSubClient;
use Google\Cloud\PubSub\Subscription;
use Google\Cloud\PubSub\Topic;
use Illuminate\Container\Container;
use Illuminate\Contracts\Encryption\Encrypter;
use Illuminate\Contracts\Queue\Job;
use Illuminate\Contracts\Queue\Queue as QueueContract;
use Illuminate\Queue\Queue;
use Illuminate\Support\Arr;
use Illuminate\Support\Facades\Log;
use RuntimeException;

class PubSubQueue extends Queue implements QueueContract
{
    use InteractsWithPubSub;

    /**
     * The Google Cloud Pub/Sub max ack seconds
     */
    const MAX_ACK_SECONDS = 10 * 60;

    /**
     * Max guaranteed message life on Google Cloud Pub/Sub
     */
    const REPUBLISH_AFTER = 5 * 24 * 60 * 60;

    /**
     * The IoC container instance.
     *
     * @var Container
     */
    protected $container;

    /**
     * The encrypter implementation.
     *
     * @var Encrypter
     */
    protected $encrypter;

    /**
     * The connection name for the queue.
     *
     * @var string
     */
    protected $connectionName;

    /**
     * @var PubSubClient The Google Cloud Pub/Sub Client.
     */
    private $pubSubClient;

    /**
     * @var string The topic:subscription to pull messages from.
     */
    private $defaultQueue;

    /**
     * @var int This value is the maximum time after a subscriber receives a message before the subscriber should
     * acknowledge the message. After message delivery but before the ack deadline expires and before the
     * message is acknowledged, it is an outstanding message and will not be delivered again during
     * that time (on a best-effort basis).
     */
    private $defaultAckDeadline;

    /**
     * PubSubQueue constructor.
     *
     * @param  array  $config
     */
    public function __construct(array $config)
    {
        $this->pubSubClient = new PubSubClient($config);
        $this->defaultQueue = data_get($config, 'queue', 'laravel');
        $this->defaultAckDeadline = data_get($config, 'retry_after', 120);
    }

    /**
     * Get the size of the queue.
     *
     * @param  string  $queue
     * @return int
     */
    public function size($queue = null)
    {
        $this->logWarning('Method `size` not implemented. Returning `1`');
        return 1;
    }

    /**
     * Push a new job onto the queue.
     *
     * @param  string  $job
     * @param  mixed  $data
     * @param  string  $queue
     * @return string
     */
    public function push($job, $data = '', $queue = null): string
    {
        return $this->pushRaw($this->constructPayload($job, $data), $queue);
    }

    /**
     * Push a raw payload onto the queue.
     *
     * @param  string  $payload
     * @param  string  $queue
     * @param  array  $options
     * @return string First message ID returned by pushBulkRaw
     */
    public function pushRaw($payload, $queue = null, array $options = []): string
    {
        return $this->pushBulkRaw([$payload], $queue, $options)[0];
    }

    /**
     * Push a raw payload onto the queue.
     *
     * @param  array  $payloads
     * @param  string  $queue
     * @param  array  $options
     * @return array
     */
    public function pushBulkRaw(array $payloads, $queue = null, array $options = []): array
    {
        $topic = $this->getTopic($queue);
        $messages = array_map(function ($payload) {
            return [
              'data' => $this->constructRawPayload($payload),
              'attributes' => null,
            ];
        }, $payloads);
        $response = $topic->publishBatch($messages);
        return $response['messageIds'];
    }

    /**
     * Push a new job onto the queue after a delay.
     *
     * @param  DateTime|int  $delay
     * @param  string  $job
     * @param  mixed  $data
     * @param  string  $queue
     * @return string
     */
    public function later($delay, $job, $data = '', $queue = null): string
    {
        return $this->pushRaw(
            $this->constructPayload($job, $data, ['availableAt' => $this->availableAt($delay)]),
            $queue
        );
    }

    /**
     * Pop the next job off of the queue.
     *
     * @param  string  $queue
     * @return Job|null
     * @throws Exception
     */
    public function pop($queue = null): ?Job
    {
        $queue = $this->getQueue($queue);
        $subscription = $this->getSubscription($queue);

        // in order to handle delayed messages we need to pull it, but not actually work it
        // laravel delays polling when it think's the queue is empty, as it would if we delayed the job and returned null
        // to avoid this, we'll loop here until we get a job that needs to actually be worked or the queue is actually empty
        while (true) {
            $messages = $subscription->pull([
              'maxMessages' => 1,
              'returnImmediately' => true,
            ]);
            if (count($messages) > 0) {
                $rawMessage = $messages[0];
                $data = $rawMessage->data();

                // deconstruct the payload
                try {
                    $payload = $this->deconstructPayload($data);
                    if (!Arr::has($payload, 'job')) {
                        throw new RuntimeException('Invalid payload');
                    }
                } catch (Exception $e) {
                    // ack(delete) bad messages, likely nothing that can be done about them
                    $subscription->acknowledge($rawMessage);
                    throw $e;
                }

                // handle delayed messages
                $availableAt = data_get($payload, 'availableAt', null);
                if ($availableAt !== null) {
                    $now = Carbon::now('UTC');
                    $secondsUntil = $this->secondsUntil(Carbon::createFromTimestampUTC($availableAt));
                    if ($secondsUntil > 0 && $secondsUntil <= self::MAX_ACK_SECONDS) {
                        // if the diff is less than the max pub/sub ack deadline, extend the deadline and leave it hanging
                        // the next time it's picked up, it'll be ready to run
                        $subscription->modifyAckDeadline($rawMessage, $secondsUntil);
                        continue;
                    } else {
                        if ($secondsUntil > self::MAX_ACK_SECONDS) {
                            if ($now->diffInSeconds(Carbon::createFromTimestampUTC($rawMessage->publishTime()->getTimestamp())) > self::REPUBLISH_AFTER) {
                                // Google deletes unacknowledged messages after 1 week
                                // once we're getting close to the deletion period, republish the message as new
                                $this->pushRaw($data, $queue);
                                $subscription->acknowledge($rawMessage);
                            } else {
                                // otherwise delay the message for the max pub/sub ack deadline and leave it hanging
                                // the next time it's picked up, it'll likely need to be delayed again
                                $subscription->modifyAckDeadline($rawMessage, self::MAX_ACK_SECONDS);
                            }
                            continue;
                        }
                    }
                }

                // set the modify ack deadline, the job now has `timeout` seconds to finish
                $timeout = data_get($payload, 'timeout') ?: $this->defaultAckDeadline;
                $timeout = $timeout > self::MAX_ACK_SECONDS ? self::MAX_ACK_SECONDS : $timeout;
                $subscription->modifyAckDeadline($rawMessage, $timeout);

                return new PubSubQueueJob($this, $this->container, $subscription, $queue, $rawMessage, $payload);
            }
            return null;
        }
    }

    /**
     * Push a new job onto the queue.
     *
     * @param  string  $queue
     * @param  string  $job
     * @param  mixed  $data
     * @return string
     */
    public function pushOn($queue, $job, $data = ''): string
    {
        return $this->push($job, $data, $queue);
    }

    /**
     * Push a new job onto the queue after a delay.
     *
     * @param  string  $queue
     * @param  DateTime|int  $delay
     * @param  string  $job
     * @param  mixed  $data
     * @return string
     */
    public function laterOn($queue, $delay, $job, $data = ''): string
    {
        return $this->later($delay, $job, $data, $queue);
    }

    /**
     * Push an array of jobs onto the queue.
     *
     * @param  array  $jobs
     * @param  mixed  $data
     * @param  string  $queue
     * @return array
     */
    public function bulk($jobs, $data = '', $queue = null): array
    {
        $payloads = array_map(function ($job) use ($data) {
            return $this->constructPayload($job, $data);
        }, (array) $jobs);
        return $this->pushBulkRaw($payloads, $queue);
    }

    /**
     * Get the connection name for the queue.
     *
     * @return string
     */
    public function getConnectionName(): string
    {
        return $this->connectionName;
    }

    /**
     * Set the connection name for the queue.
     *
     * @param  string  $name
     * @return $this
     */
    public function setConnectionName($name): self
    {
        $this->connectionName = $name;

        return $this;
    }

    /**
     * Set the IoC container instance.
     *
     * @param  Container  $container
     */
    public function setContainer(Container $container): void
    {
        $this->container = $container;
    }

    /**
     * Set the encrypter
     *
     * @param  Encrypter  $encrypter
     */
    public function setEncrypter(Encrypter $encrypter): void
    {
        $this->encrypter = $encrypter;
    }

    /**
     * Get the queue or default
     *
     * @param string $queue
     * @return string
     */
    private function getQueue($queue): string
    {
        return !empty($queue) ? $queue : $this->defaultQueue;
    }

    /**
     * @param  string  $queue
     * @return Topic
     */
    private function getTopic($queue): Topic
    {
        $topic = $this->getQueue($queue);
        $parts = preg_split('/:/', $queue, 2);
        if (count($parts) > 1) {
            $topic = $parts[0];
        }
        return $this->pubSubClient->topic($topic);
    }

    /**
     * Get the subscription
     *
     * @param  string|null  $queue  The queue/topic name
     * @return Subscription
     */
    private function getSubscription($queue = null): Subscription
    {
        $topic = $this->getQueue($queue);
        $subscription = $topic;
        $parts = preg_split('/:/', $topic, 2);
        if (count($parts) > 1) {
            $topic = $parts[0];
            $subscription = $parts[1];
        }
        return $this->pubSubClient->subscription($subscription, $topic);
    }

    /**
     * Log a warning message.
     *
     * @param  string  $message
     * @param  array  $context
     */
    private function logWarning($message, array $context = []): void
    {
        Log::warning('Google Pub/Sub Queue Driver: '.$message, $context);
    }

    /**
     * Get the number of seconds until the given DateTime.
     *
     * @param  DateTimeInterface  $delay
     * @return int
     */
    protected function secondsUntil($delay): int
    {
        return $delay instanceof DateTimeInterface
          ? max(0, $delay->getTimestamp() - $this->currentTime())
          : (int) $delay;
    }

    /**
     * Get the "available at" UNIX timestamp.
     *
     * @param  DateTimeInterface|int  $delay
     * @return int
     */
    protected function availableAt($delay = 0): int
    {
        return $delay instanceof DateTimeInterface
          ? $delay->getTimestamp()
          : Carbon::now()->addSeconds($delay)->getTimestamp();
    }

    /**
     * Get the current system time as a UNIX timestamp.
     *
     * @return int
     */
    protected function currentTime(): int
    {
        return Carbon::now()->getTimestamp();
    }

    /**
     * Consume all of the messages on the queue
     *
     * @param  String  $queue  The queue to consume
     * @return int Number of messages consumed
     */
    public function consume($queue): int
    {
        $queue = $this->getQueue($queue);
        $subscription = $this->getSubscription($queue);

        $consumed = 0;
        while (true) {
            $messages = $subscription->pull([
              'maxMessages' => 100,
              'returnImmediately' => true,
            ]);

            $count = count($messages);
            if ($count <= 0) {
                break;
            }

            $subscription->acknowledgeBatch($messages);

            $consumed += $count;
        }

        return $consumed;
    }
}
